/* Solar system plotting programs, for use with nb1d*.cs solar system	*/
/* simulations (any of the 1D programs).  Solution shape may be		*/
/* either 2D or 3D, programs which use these must help with		*/
/* initialization of the plotter.					*/

#include <stdio.h>
#include <cscomm.h>
#include "colorp.h"

extern shape pair;	/* 2D or 3D */

/* Image representation and transformation data */

shape [1024][512]screen;	/*  The screen space, we use zoom 1 1	*/

bool:screen xcntr;		/* pixel is center of object */

short:screen red, grn, blu,	/* color values */
    redtrl, grntrl, blutrl;	/* Trails left by objects */

int:screen rgb;			/* Combined rgb */

float pcx=0.0,		/* View center - x		*/
      pcy=0.0,		/* View center - y		*/
      scrnht=27.305,	/* Physical screen height (cm)	*/
      scrnwd=34.29,	/* Physical screen width (cm)	*/
      xwidth=60.00,	/* Width of view, space units (60.00 gets Neptune) */
      xscale=1.0,	/* Scale factor for x		*/
      yscale=1.0;	/* Scale factor for y		*/

int plft=0,	/* left limit of view (screen coordinates)	*/
    prgt=639,	/* right limit of view (screen coordinates)	*/
    ptop=0,	/* top limit of view (screen coordinates)	*/
    pbot=511,	/* bottom limit of view (screen coordinates)	*/
    pcv=256,	/* position of center in screen (row)		*/
    pch=320,	/* position of center in screen (col)		*/
    drgb=1;	/* intensity decrement for trails		*/

/**** Limits are inclusive ****/

/* Icons */

short:pair ired, igrn, iblu,	/* r,g,b for icons */
	   icx, icy;		/* pixel offset	   */

bool:pair iactv,	/* Part of icon		*/
	  icntr;	/* Center of icon	*/

/* Clear image to black */

void clrimage()
 { with (screen)
   { red = 0;  grn = 0;  blu = 0;  xcntr = 0;
     redtrl = 0;  grntrl = 0;  blutrl = 0;
   }
 }

/* Project positions in xp.  Call with pair current, processors */
/* containing current positions active.				*/

void project(float:current xp[3])
 {int:current prp, pcp;		/* Pixel coordinates	*/
  CM_sendaddr_t:current ptg;	/* Pixel address	*/
  with (screen) { red = 0;  grn = 0;  blu = 0;  xcntr = 0; }
  where (iactv)
   { prp = (int:current)((xp[1] - pcy)*yscale) + icy + pcv;
     pcp = (int:current)((xp[0] - pcx)*xscale) + icx + pch;
     where ( iactv & (ptop <= prp) & (prp <= pbot)
		   & (plft <= pcp) & (pcp <= prgt) )
      { ptg = make_send_address(screen, pcp, prp);
	send(&red, ptg, ired, CMC_combiner_add, CMC_no_field);
	send(&grn, ptg, igrn, CMC_combiner_add, CMC_no_field);
	send(&blu, ptg, iblu, CMC_combiner_add, CMC_no_field);
	where (icntr)
	 send(&xcntr, ptg, icntr, CMC_combiner_logior, CMC_no_field);
      }
   }
 }

/* Construct final image with orbit trails, decrements trail	*/
/* intensities by dlum.						*/

void imagesol(int dlum)
 { with (screen)
   {/* Clip intensities */
    where ( red > 255 ) red = 255;  where ( red < 0 ) red = 0;
    where ( grn > 255 ) grn = 255;  where ( grn < 0 ) grn = 0;
    where ( blu > 255 ) blu = 255;  where ( blu < 0 ) blu = 0;
    /* Decrement trail intensities */
    redtrl = (redtrl - dlum) >? 0;
    grntrl = (grntrl - dlum) >? 0;
    blutrl = (blutrl - dlum) >? 0;
    /* Insert new trail points, add old trails to image */
    where ( xcntr ) { redtrl = red;  grntrl = grn;  blutrl = blu; }
    where ( red == 0 ) red = redtrl;
    where ( grn == 0 ) grn = grntrl;
    where ( blu == 0 ) blu = blutrl;
   }
 }

void setup_proj(float cx, float cy, float xw)
 { pcx = cx;  pcy = cy;  xwidth = xw;
   xscale = (prgt - plft)/xwidth;
   yscale = (ptop - pbot)*scrnwd/(xwidth*scrnht);
   clrimage();
 }

void plot_planets(float:pair xp[3])
 { project(xp);
   with (screen) { imagesol(drgb);  sc(); }
 }

/**********************************************************************

Icon pixels are represented by triples

int rx, ry:  relative location
float   xi:  relative intensity

Icon pixels are generally given brightest pixels first so that if the
geometry isn't big enough to store everything we can keep the most
important parts.

***********************************************************************/

struct iplx { int rx, ry; float xi; };

/* Disk of diameter 2 pixels*/

#define ic2n 9

struct iplx ic2[ic2n] =
 { { 0,  0, 1.00},
   { 0, -1, 0.44}, {-1,  0, 0.44}, { 1,  0, 0.44}, { 0, 1, 0.44},
   {-1, -1, 0.08}, { 1, -1, 0.08}, {-1,  1, 0.08}, { 1, 1, 0.08} };

/* Disk of diameter 3 pixels */

#define ic3n 9

struct iplx ic3[ic3n] =
 { { 0,  0, 1.00},
   { 0, -1, 0.97}, {-1,  0, 0.97}, { 1,  0, 0.97}, { 0, 1, 0.97},
   {-1, -1, 0.44}, { 1, -1, 0.44}, {-1,  1, 0.44}, { 1, 1, 0.44} };

/* Disk of diameter 4 pixels */

#define ic4n 21

struct iplx ic4[ic4n] =
 { { 0,  0, 1.00},
   { 0, -1, 1.00}, {-1,  0, 1.00}, { 1,  0, 1.00}, { 0,  1, 1.00},
   {-1, -1, 0.97}, { 1, -1, 0.97}, {-1,  1, 0.97}, { 1,  1, 0.97},
   { 0, -2, 0.44}, {-2,  0, 0.44}, { 2,  0, 0.44}, { 0,  2, 0.44},
   {-1, -2, 0.24}, { 1, -2, 0.24}, {-2, -1, 0.24}, { 2, -1, 0.24},
   {-2,  1, 0.24}, { 2,  1, 0.24}, {-1,  2, 0.24}, { 1,  2, 0.24} };

/* Disk of diameter 5 pixels */

#define ic5n 25

struct iplx ic5[ic5n] =
 { { 0,  0, 1.00},
   {-1, -1, 1.00}, { 0, -1, 1.00}, { 1, -1, 1.00}, {-1,  0, 1.00},
   { 1,  0, 1.00}, {-1,  1, 1.00}, { 0,  1, 1.00}, { 1,  1, 1.00},
   { 0, -2, 0.98}, {-2,  0, 0.98}, { 2,  0, 0.98}, { 0,  2, 0.98},
   {-1, -2, 0.75}, { 1, -2, 0.75}, {-2, -1, 0.75}, { 2, -1, 0.75},
   {-2,  1, 0.75}, { 2,  1, 0.75}, {-1,  2, 0.75}, { 1,  2, 0.75},
   {-2, -2, 0.12}, { 2, -2, 0.12}, {-2,  2, 0.12}, { 2,  2, 0.12} };

/* Disk of diameter 6 pixels */

#define ic6n 37

struct iplx ic6[ic6n] =
 { { 0, 0,  1.00},
   {-1, -2, 1.00}, { 0, -2, 1.00}, { 1, -2, 1.00},
   {-2, -1, 1.00}, {-1, -1, 1.00}, { 0, -1, 1.00}, { 1, -1, 1.00}, { 2, -1, 1.00},
   {-2,  0, 1.00}, {-1,  0, 1.00}, { 1,  0, 1.00}, { 2,  0, 1.00},
   {-2,  1, 1.00}, {-1,  1, 1.00}, { 0,  1, 1.00}, { 1,  1, 1.00}, { 2,  1, 1.00},
   {-1,  2, 1.00}, { 0,  2, 1.00}, { 1,  2, 1.00},
   {-2, -2, 0.67}, { 2, -2, 0.67}, {-2,  2, 0.67}, { 2,  2, 0.67},
   { 0, -3, 0.48}, {-3,  0, 0.48}, { 3,  0, 0.48}, { 0,  3, 0.48},
   {-1, -3, 0.25}, { 1, -3, 0.25}, {-3, -1, 0.25}, { 3, -1, 0.25},
   {-3,  1, 0.25}, { 3,  1, 0.25}, {-1,  3, 0.25}, { 1,  3, 0.25} };

/* Special for Saturn, disk of diameter 5 pixels with rings */

#define ic5rn 37

struct iplx ic5r[ic5rn] =
 { { 0,  0, 1.00},
   {-1, -1, 1.00}, { 0, -1, 1.00}, { 1, -1, 1.00}, {-1,  0, 1.00},
   { 1,  0, 1.00}, {-1,  1, 1.00}, { 0,  1, 1.00}, { 1,  1, 1.00},
   { 0, -2, 0.98}, {-2,  0, 0.98}, { 2,  0, 0.98}, { 0,  2, 0.98},
   {-2, -1, 0.90}, { 2, -1, 0.90}, {-2,  1, 0.90}, { 2,  1, 0.90},
   {-3, -1, 0.80}, { 3, -1, 0.80}, {-3,  1, 0.80}, { 3,  1, 0.80},
   {-4,  0, 0.80}, { 4,  0, 0.80},
   {-1, -2, 0.75}, { 1, -2, 0.75}, {-1,  2, 0.75}, { 1,  2, 0.75},
   {-2, -2, 0.12}, { 2, -2, 0.12}, {-2,  2, 0.12}, { 2,  2, 0.12},
   {-4, -1, 0.10}, { 4, -1, 0.10}, {-3,  0, 0.10}, { 3,  0, 0.10},
   {-4,  1, 0.10}, { 4,  1, 0.10} };

/* Function to print icon for inspection.  Offsets are required	   */
/* to satisfy -10 <= offset <= 10 in both axes, but the program	   */
/* is unsatisfactory when the total width is more than 15 pixels.  */

void print_ic( int in, struct iplx *ics )
 {int xl=0, xr=0, yt=0, yb=0, ix, iy, ik;
  float fx, xx[21][21];  struct iplx *ip;
  /* Zero icon array */
  for ( ix = 0; ix < 21; ix++ )
   for ( iy = 0; iy < 21;  iy++ )  xx[ix][iy] = 0.0;
  /* Find limits, fill in array */
  for ( ik = 0; ik < in; ik++ )
   { ip = &ics[ik];  ix = ip->rx;  iy = ip->ry;  fx = ip->xi;
     xx[10+ix][10+iy] = fx;
     if ( ix < xl ) xl = ix; else if ( ix > xr ) xr = ix;
     if ( iy < yt ) yt = iy; else if ( iy > yb ) yb = iy;
   }
  /* Print the icon */
  printf("The icon has %d pixels.\n  ", in);
  for ( ix = xl; ix <= xr; ix++ ) printf("  %3d", ix);
  for ( iy = yt; iy <= yb; iy++ )
   { printf("\n%3d ", iy);
     for ( ix = xl; ix <= xr;  ix++ ) printf(" %4.2f", xx[10+ix][10+iy]);
   }
 }

/* Load icon icp  with `in' pixels into CM with color (ir,ig,ib).	*/
/* Call in `pair' shape with all processors for this icon active.	*/
/* j is the coordinate pvar for redundant representations;  valid for	*/
/* 0 <= j < jmx .							*/

void load_icon( int in, struct iplx icp[], int ir, int ig, int ib,
		int:current *j, int jmx )
 { int ij;  float:current pxf;  struct iplx *ip;
   iactv = 0;  icntr = 0;   ired = ir;  igrn = ig;  iblu = ib;
   jmx <?= in;	/* Loop limit */
   for ( ij = 0;  ij < jmx;  ij++ )
    { ip = &icp[ij];
      where ( *j == ij )
       { iactv = 1;  icx = ip->rx;  icy = ip->ry;  pxf = ip->xi; }
    }
   where ( iactv )
    { ired *= pxf;  igrn *= pxf;  iblu *= pxf;
      where ( icx == 0 & icy == 0 ) icntr = 1;
    }
 }

/* Load solar system icons.  *i is body coordinate, *j as for load_icon.  */
/* Order of bodies is Earth, Mars, Jupiter, Saturn, Uranus, Neptune, Sun, */
/* all the rest asteroids.  Call with all processors active.		  */
/* Valid coordinates are 0 <= j < jmx .					  */

void load_solar( int:current *i, int:current *j, int jmx )
 { where ( *i == 0 ) load_icon(ic3n,  ic3,  170, 170, 255, j, jmx);
   where ( *i == 1 ) load_icon(ic3n,  ic3,  255, 150, 150, j, jmx);
   where ( *i == 2 ) load_icon(ic5n,  ic5,  240, 255, 100, j, jmx);
   where ( *i == 3 ) load_icon(ic5rn, ic5r, 255, 200, 100, j, jmx);
   where ( *i == 4 ) load_icon(ic5n,  ic5,  170, 255, 170, j, jmx);
   where ( *i == 5 ) load_icon(ic5n,  ic5,  128, 128, 255, j, jmx);
   where ( *i == 6 ) load_icon(ic6n,  ic6,  255, 255, 255, j, jmx);
   where ( *i >= 7 ) load_icon(ic2n,  ic2,  255, 255, 220, j, jmx);
 }
